/*

  Broadcom B43 wireless driver

  debugfs driver debugging code

  Copyright (c) 2005 Michael Buesch <mb@bu3sch.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 51 Franklin Steet, Fifth Floor,
  Boston, MA 02110-1301, USA.

*/

#include <linux/fs.h>
#include <linux/debugfs.h>
#include <linux/slab.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/mutex.h>

#include "b43.h"
#include "main.h"
#include "debugfs.h"
#include "dma.h"
#include "pio.h"
#include "xmit.h"

#define REALLY_BIG_BUFFER_SIZE	(1024*256)

static struct b43_debugfs fs;
static char big_buffer[1024 * 256];
static DEFINE_MUTEX(big_buffer_mutex);

static ssize_t write_file_dummy(struct file *file, const char __user * buf,
				size_t count, loff_t * ppos)
{
	return count;
}

static int open_file_generic(struct inode *inode, struct file *file)
{
	file->private_data = inode->i_private;
	return 0;
}

#define fappend(fmt, x...)	pos += snprintf(buf + pos, len - pos, fmt , ##x)

static ssize_t drvinfo_read_file(struct file *file, char __user * userbuf,
				 size_t count, loff_t * ppos)
{
	const size_t len = ARRAY_SIZE(big_buffer);
	char *buf = big_buffer;
	size_t pos = 0;
	ssize_t res;

	mutex_lock(&big_buffer_mutex);
	/* This is where the information is written to the "driver" file */
	fappend(KBUILD_MODNAME " driver\n");
	fappend("Compiled at: %s %s\n", __DATE__, __TIME__);
	res = simple_read_from_buffer(userbuf, count, ppos, buf, pos);
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t tsf_read_file(struct file *file, char __user * userbuf,
			     size_t count, loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	const size_t len = ARRAY_SIZE(big_buffer);
	char *buf = big_buffer;
	size_t pos = 0;
	ssize_t res;
	unsigned long flags;
	u64 tsf;

	mutex_lock(&big_buffer_mutex);
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_STARTED) {
		fappend("Board not initialized.\n");
		goto out;
	}
	b43_tsf_read(dev, &tsf);
	fappend("0x%08x%08x\n",
		(unsigned int)((tsf & 0xFFFFFFFF00000000ULL) >> 32),
		(unsigned int)(tsf & 0xFFFFFFFFULL));

      out:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
	res = simple_read_from_buffer(userbuf, count, ppos, buf, pos);
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t tsf_write_file(struct file *file, const char __user * user_buf,
			      size_t count, loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	char *buf = big_buffer;
	ssize_t buf_size;
	ssize_t res;
	unsigned long flags;
	u64 tsf;

	mutex_lock(&big_buffer_mutex);
	buf_size = min(count, ARRAY_SIZE(big_buffer) - 1);
	if (copy_from_user(buf, user_buf, buf_size)) {
		res = -EFAULT;
		goto out_unlock_bb;
	}
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_STARTED) {
		b43err(dev->wl, "debugfs: Board not initialized.\n");
		res = -EFAULT;
		goto out_unlock;
	}
	if (sscanf(buf, "%llu", (unsigned long long *)(&tsf)) != 1) {
		b43err(dev->wl, "debugfs: invalid values for \"tsf\"\n");
		res = -EINVAL;
		goto out_unlock;
	}
	b43_tsf_write(dev, tsf);
	mmiowb();
	res = buf_size;

      out_unlock:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
      out_unlock_bb:
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t txstat_read_file(struct file *file, char __user * userbuf,
				size_t count, loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	struct b43_dfsentry *e = dev->dfsentry;
	struct b43_txstatus_log *log = &e->txstatlog;
	unsigned long flags;
	char *buf = log->printbuf;
	const size_t len = ARRAY_SIZE(log->printbuf);
	size_t pos = 0;
	ssize_t res;
	int i, idx;
	struct b43_txstatus *stat;

	mutex_lock(&big_buffer_mutex);
	spin_lock_irqsave(&log->lock, flags);
	if (!log->printing) {
		log->printing = 1;
		fappend("b43 TX status reports:\n\n"
			"index | cookie | seq | phy_stat | frame_count | "
			"rts_count | supp_reason | pm_indicated | "
			"intermediate | for_ampdu | acked\n" "---\n");
		i = log->end + 1;
		idx = 0;
		while (1) {
			if (log->end < 0) {
				fappend("Nothing transmitted, yet\n");
				break;
			}
			if (i == B43_NR_LOGGED_TXSTATUS)
				i = 0;
			stat = &(log->log[i]);
			if (stat->cookie) {
				fappend("%03d | "
					"0x%04X | 0x%04X | 0x%02X | "
					"0x%X | 0x%X | "
					"%u | %u | "
					"%u | %u | %u\n",
					idx,
					stat->cookie, stat->seq, stat->phy_stat,
					stat->frame_count, stat->rts_count,
					stat->supp_reason, stat->pm_indicated,
					stat->intermediate, stat->for_ampdu,
					stat->acked);
				idx++;
			}
			if (i == log->end)
				break;
			i++;
		}
		log->buf_avail = pos;
	}
	memcpy(big_buffer, buf, min(log->buf_avail, ARRAY_SIZE(big_buffer)));
	spin_unlock_irqrestore(&log->lock, flags);

	res = simple_read_from_buffer(userbuf, count, ppos,
				      big_buffer, log->buf_avail);
	if (*ppos == log->buf_avail) {
		spin_lock_irqsave(&log->lock, flags);
		log->printing = 0;
		spin_unlock_irqrestore(&log->lock, flags);
	}
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t restart_write_file(struct file *file,
				  const char __user * user_buf, size_t count,
				  loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	char *buf = big_buffer;
	ssize_t buf_size;
	ssize_t res;
	unsigned long flags;

	mutex_lock(&big_buffer_mutex);
	buf_size = min(count, ARRAY_SIZE(big_buffer) - 1);
	if (copy_from_user(buf, user_buf, buf_size)) {
		res = -EFAULT;
		goto out_unlock_bb;
	}
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_INITIALIZED) {
		b43err(dev->wl, "debugfs: Board not initialized.\n");
		res = -EFAULT;
		goto out_unlock;
	}
	if (count > 0 && buf[0] == '1') {
		b43_controller_restart(dev, "manually restarted");
		res = count;
	} else
		res = -EINVAL;

      out_unlock:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
      out_unlock_bb:
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t txpower_g_read_file(struct file *file, char __user * userbuf,
				   size_t count, loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	const size_t len = ARRAY_SIZE(big_buffer);
	char *buf = big_buffer;
	size_t pos = 0;
	ssize_t res;
	unsigned long flags;

	mutex_lock(&big_buffer_mutex);
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_STARTED) {
		fappend("Not initialized\n");
		goto out;
	}
	if (dev->phy.type != B43_PHYTYPE_G) {
		fappend("Device is not a G-PHY\n");
		goto out;
	}
	fappend("Control:               %s\n", dev->phy.manual_txpower_control ?
		"MANUAL" : "AUTOMATIC");
	fappend("Baseband attenuation:  %u\n", dev->phy.bbatt.att);
	fappend("Radio attenuation:     %u\n", dev->phy.rfatt.att);
	fappend("TX Mixer Gain:         %s\n",
		(dev->phy.tx_control & B43_TXCTL_TXMIX) ? "ON" : "OFF");
	fappend("PA Gain 2dB:           %s\n",
		(dev->phy.tx_control & B43_TXCTL_PA2DB) ? "ON" : "OFF");
	fappend("PA Gain 3dB:           %s\n",
		(dev->phy.tx_control & B43_TXCTL_PA3DB) ? "ON" : "OFF");
	fappend("\n\n");
	fappend("You can write to this file:\n");
	fappend("Writing \"auto\" enables automatic txpower control.\n");
	fappend
	    ("Writing the attenuation values as \"bbatt rfatt txmix pa2db pa3db\" "
	     "enables manual txpower control.\n");
	fappend("Example: 5 4 0 0 1\n");
	fappend("Enables manual control with Baseband attenuation 5, "
		"Radio attenuation 4, No TX Mixer Gain, "
		"No PA Gain 2dB, With PA Gain 3dB.\n");

      out:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
	res = simple_read_from_buffer(userbuf, count, ppos, buf, pos);
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static ssize_t txpower_g_write_file(struct file *file,
				    const char __user * user_buf, size_t count,
				    loff_t * ppos)
{
	struct b43_wldev *dev = file->private_data;
	char *buf = big_buffer;
	ssize_t buf_size;
	ssize_t res;
	unsigned long flags, phy_flags;

	mutex_lock(&big_buffer_mutex);
	buf_size = min(count, ARRAY_SIZE(big_buffer) - 1);
	if (copy_from_user(buf, user_buf, buf_size)) {
		res = -EFAULT;
		goto out_unlock_bb;
	}
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_STARTED) {
		b43err(dev->wl, "debugfs: Board not initialized.\n");
		res = -ENODEV;
		goto out_unlock;
	}
	if (dev->phy.type != B43_PHYTYPE_G) {
		b43err(dev->wl, "debugfs: Device is not a G-PHY\n");
		res = -ENODEV;
		goto out_unlock;
	}
	if ((buf_size >= 4) && (memcmp(buf, "auto", 4) == 0)) {
		/* Automatic control */
		dev->phy.manual_txpower_control = 0;
		b43_phy_xmitpower(dev);
	} else {
		int bbatt = 0, rfatt = 0, txmix = 0, pa2db = 0, pa3db = 0;
		/* Manual control */
		if (sscanf(buf, "%d %d %d %d %d", &bbatt, &rfatt,
			   &txmix, &pa2db, &pa3db) != 5) {
			b43err(dev->wl,
			       "debugfs: invalid value for \"tx_power_g\"\n");
			res = -EINVAL;
			goto out_unlock;
		}
		b43_put_attenuation_into_ranges(dev, &bbatt, &rfatt);
		dev->phy.manual_txpower_control = 1;
		dev->phy.bbatt.att = bbatt;
		dev->phy.rfatt.att = rfatt;
		dev->phy.tx_control = 0;
		if (txmix)
			dev->phy.tx_control |= B43_TXCTL_TXMIX;
		if (pa2db)
			dev->phy.tx_control |= B43_TXCTL_PA2DB;
		if (pa3db)
			dev->phy.tx_control |= B43_TXCTL_PA3DB;
		b43_phy_lock(dev, phy_flags);
		b43_radio_lock(dev);
		b43_set_txpower_g(dev, &dev->phy.bbatt,
				  &dev->phy.rfatt, dev->phy.tx_control);
		b43_radio_unlock(dev);
		b43_phy_unlock(dev, phy_flags);
	}
	res = buf_size;
      out_unlock:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
      out_unlock_bb:
	mutex_unlock(&big_buffer_mutex);

	return res;
}

static size_t append_lo_table(size_t pos, char *buf, const size_t len,
			      struct b43_loctl table[B43_NR_BB][B43_NR_RF])
{
	unsigned int i, j;
	struct b43_loctl *ctl;

	for (i = 0; i < B43_NR_BB; i++) {
		for (j = 0; j < B43_NR_RF; j++) {
			ctl = &(table[i][j]);
			fappend("(bbatt %2u, rfatt %2u)  ->  "
				"(I %+3d, Q %+3d, Used: %d, Calibrated: %d)\n",
				i, j, ctl->i, ctl->q,
				ctl->used,
				b43_loctl_is_calibrated(ctl));
		}
	}

	return pos;
}

static ssize_t loctls_read_file(struct file *file, char __user *userbuf,
				size_t count, loff_t *ppos)
{
	struct b43_wldev *dev = file->private_data;
	const size_t len = ARRAY_SIZE(big_buffer);
	char *buf = big_buffer;
	size_t pos = 0;
	ssize_t res;
	unsigned long flags;
	struct b43_txpower_lo_control *lo;
	unsigned int i;

	mutex_lock(&big_buffer_mutex);
	mutex_lock(&dev->wl->mutex);
	spin_lock_irqsave(&dev->wl->irq_lock, flags);
	if (b43_status(dev) < B43_STAT_INITIALIZED) {
		fappend("Not initialized\n");
		goto out;
	}
	if (dev->phy.type != B43_PHYTYPE_G) {
		fappend("Device is not a G-PHY\n");
		goto out;
	}

	lo = dev->phy.lo_control;
	fappend("-- Local Oscillator calibration data --\n\n");
	fappend("Measured: %d,  Rebuild: %d,  HW-power-control: %d\n",
		lo->lo_measured,
		lo->rebuild,
		dev->phy.hardware_power_control);
	fappend("TX Bias: 0x%02X,  TX Magn: 0x%02X\n",
		lo->tx_bias, lo->tx_magn);
	fappend("Power Vector: 0x%08X%08X\n",
		(unsigned int)((lo->power_vector & 0xFFFFFFFF00000000ULL) >> 32),
		(unsigned int)(lo->power_vector & 0x00000000FFFFFFFFULL));
	fappend("\nControl table WITH PADMIX:\n");
	pos = append_lo_table(pos, buf, len, lo->with_padmix);
	fappend("\nControl table WITHOUT PADMIX:\n");
	pos = append_lo_table(pos, buf, len, lo->no_padmix);
	fappend("\nUsed RF attenuation values:  Value(WithPadmix flag)\n");
	for (i = 0; i < lo->rfatt_list.len; i++) {
		fappend("%u(%d), ",
			lo->rfatt_list.list[i].att,
			lo->rfatt_list.list[i].with_padmix);
	}
	fappend("\n");
	fappend("\nUsed Baseband attenuation values:\n");
	for (i = 0; i < lo->bbatt_list.len; i++) {
		fappend("%u, ",
			lo->bbatt_list.list[i].att);
	}
	fappend("\n");

out:
	spin_unlock_irqrestore(&dev->wl->irq_lock, flags);
	mutex_unlock(&dev->wl->mutex);
	res = simple_read_from_buffer(userbuf, count, ppos, buf, pos);
	mutex_unlock(&big_buffer_mutex);

	return res;
}

#undef fappend

static struct file_operations drvinfo_fops = {
	.read = drvinfo_read_file,
	.write = write_file_dummy,
	.open = open_file_generic,
};

static struct file_operations tsf_fops = {
	.read = tsf_read_file,
	.write = tsf_write_file,
	.open = open_file_generic,
};

static struct file_operations txstat_fops = {
	.read = txstat_read_file,
	.write = write_file_dummy,
	.open = open_file_generic,
};

static struct file_operations txpower_g_fops = {
	.read = txpower_g_read_file,
	.write = txpower_g_write_file,
	.open = open_file_generic,
};

static struct file_operations restart_fops = {
	.write = restart_write_file,
	.open = open_file_generic,
};

static struct file_operations loctls_fops = {
	.read = loctls_read_file,
	.open = open_file_generic,
};


int b43_debug(struct b43_wldev *dev, enum b43_dyndbg feature)
{
	return !!(dev->dfsentry && dev->dfsentry->dyn_debug[feature]);
}

static void b43_remove_dynamic_debug(struct b43_wldev *dev)
{
	struct b43_dfsentry *e = dev->dfsentry;
	int i;

	for (i = 0; i < __B43_NR_DYNDBG; i++)
		debugfs_remove(e->dyn_debug_dentries[i]);
}

static void b43_add_dynamic_debug(struct b43_wldev *dev)
{
	struct b43_dfsentry *e = dev->dfsentry;
	struct dentry *d;

#define add_dyn_dbg(name, id, initstate) do {		\
	e->dyn_debug[id] = (initstate);			\
	d = debugfs_create_bool(name, 0600, e->subdir,	\
				&(e->dyn_debug[id]));	\
	if (!IS_ERR(d))					\
		e->dyn_debug_dentries[id] = d;		\
				} while (0)

	add_dyn_dbg("debug_xmitpower", B43_DBG_XMITPOWER, 0);
	add_dyn_dbg("debug_dmaoverflow", B43_DBG_DMAOVERFLOW, 0);
	add_dyn_dbg("debug_dmaverbose", B43_DBG_DMAVERBOSE, 0);
	add_dyn_dbg("debug_pwork_fast", B43_DBG_PWORK_FAST, 0);
	add_dyn_dbg("debug_pwork_stop", B43_DBG_PWORK_STOP, 0);

#undef add_dyn_dbg
}

void b43_debugfs_add_device(struct b43_wldev *dev)
{
	struct b43_dfsentry *e;
	struct b43_txstatus_log *log;
	char devdir[16];

	B43_WARN_ON(!dev);
	e = kzalloc(sizeof(*e), GFP_KERNEL);
	if (!e) {
		b43err(dev->wl, "debugfs: add device OOM\n");
		return;
	}
	e->dev = dev;
	log = &e->txstatlog;
	log->log = kcalloc(B43_NR_LOGGED_TXSTATUS,
			   sizeof(struct b43_txstatus), GFP_KERNEL);
	if (!log->log) {
		b43err(dev->wl, "debugfs: add device txstatus OOM\n");
		kfree(e);
		return;
	}
	log->end = -1;
	spin_lock_init(&log->lock);

	dev->dfsentry = e;

	snprintf(devdir, sizeof(devdir), "%s", wiphy_name(dev->wl->hw->wiphy));
	e->subdir = debugfs_create_dir(devdir, fs.root);
	if (!e->subdir || IS_ERR(e->subdir)) {
		if (e->subdir == ERR_PTR(-ENODEV)) {
			b43dbg(dev->wl, "DebugFS (CONFIG_DEBUG_FS) not "
			       "enabled in kernel config\n");
		} else {
			b43err(dev->wl, "debugfs: cannot create %s directory\n",
			       devdir);
		}
		dev->dfsentry = NULL;
		kfree(log->log);
		kfree(e);
		return;
	}

	e->dentry_tsf = debugfs_create_file("tsf", 0600, e->subdir,
					    dev, &tsf_fops);
	if (IS_ERR(e->dentry_tsf))
		e->dentry_tsf = NULL;
	e->dentry_txstat = debugfs_create_file("tx_status", 0400, e->subdir,
					       dev, &txstat_fops);
	if (IS_ERR(e->dentry_txstat))
		e->dentry_txstat = NULL;
	e->dentry_txpower_g = debugfs_create_file("tx_power_g", 0600, e->subdir,
						  dev, &txpower_g_fops);
	if (IS_ERR(e->dentry_txpower_g))
		e->dentry_txpower_g = NULL;
	e->dentry_restart = debugfs_create_file("restart", 0200, e->subdir,
						dev, &restart_fops);
	if (IS_ERR(e->dentry_restart))
		e->dentry_restart = NULL;
	e->dentry_loctls = debugfs_create_file("loctls", 0400, e->subdir,
					       dev, &loctls_fops);
	if (IS_ERR(e->dentry_loctls))
		e->dentry_loctls = NULL;

	b43_add_dynamic_debug(dev);
}

void b43_debugfs_remove_device(struct b43_wldev *dev)
{
	struct b43_dfsentry *e;

	if (!dev)
		return;
	e = dev->dfsentry;
	if (!e)
		return;
	b43_remove_dynamic_debug(dev);
	debugfs_remove(e->dentry_loctls);
	debugfs_remove(e->dentry_tsf);
	debugfs_remove(e->dentry_txstat);
	debugfs_remove(e->dentry_restart);
	debugfs_remove(e->dentry_txpower_g);
	debugfs_remove(e->subdir);
	kfree(e->txstatlog.log);
	kfree(e);
}

void b43_debugfs_log_txstat(struct b43_wldev *dev,
			    const struct b43_txstatus *status)
{
	struct b43_dfsentry *e = dev->dfsentry;
	struct b43_txstatus_log *log;
	struct b43_txstatus *cur;
	int i;

	if (!e)
		return;
	log = &e->txstatlog;
	B43_WARN_ON(!irqs_disabled());
	spin_lock(&log->lock);
	i = log->end + 1;
	if (i == B43_NR_LOGGED_TXSTATUS)
		i = 0;
	log->end = i;
	cur = &(log->log[i]);
	memcpy(cur, status, sizeof(*cur));
	spin_unlock(&log->lock);
}

void b43_debugfs_init(void)
{
	memset(&fs, 0, sizeof(fs));
	fs.root = debugfs_create_dir(KBUILD_MODNAME, NULL);
	if (!fs.root || IS_ERR(fs.root)) {
		fs.root = NULL;
		return;
	}
	fs.dentry_driverinfo = debugfs_create_file("driver", 0444, fs.root,
						   NULL, &drvinfo_fops);
	if (IS_ERR(fs.dentry_driverinfo))
		fs.dentry_driverinfo = NULL;
}

void b43_debugfs_exit(void)
{
	debugfs_remove(fs.dentry_driverinfo);
	debugfs_remove(fs.root);
}
